
#include "http_comm.h"
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <io.h>
#include <winsock2.h>
#else
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include "md5.h"
#ifdef GZIP
#include <zlib.h>
#endif



#define READBUFFER      2048
#define MAX_SOCKS       1
#define DEBUGLEVEL      5

int listen_sock;
struct sockaddr_in listen_sock_name;

fd_set fd_listen_set;
char *proxy_host = NULL;
int proxy_port = 0;

/*===========================================================================*/
/*==========                   Helper Functions                    ==========*/
/*===========================================================================*/

unsigned getFileSize(int file)
{
   unsigned cur_pos = lseek(file, 0, SEEK_CUR);
   unsigned length = lseek(file, 0, SEEK_END);
   lseek(file, cur_pos, SEEK_SET);
   return length;
}


char *getMD5(int file)
{
   struct MD5Context *context = malloc(sizeof(struct MD5Context));
   unsigned char digest[16];
   unsigned char *buf = malloc(READBUFFER);
   unsigned length = getFileSize(file);
   int nbytes = READBUFFER;
   unsigned read_bytes = 0;
   unsigned total_bytes = 0;
   char *md5 = (char *) malloc(33);
   int i = 0;

   MD5Init(context);
   do
   {
      read_bytes = read(file, buf, nbytes);
      MD5Update(context, buf, read_bytes);
      total_bytes += read_bytes;
      if (length - total_bytes < READBUFFER)
         nbytes = length - total_bytes;
   }
   while (read_bytes);
   MD5Final(digest, context);
   free(context);
   free(buf);
   for (i = 0; i < 16; i++)
      sprintf(md5 + i * 2, "%02x", digest[i]);
   md5[32] = '\0';
   return md5;
}

int getIdFromUri(char *uri)
{
   int id = -1;
   if (strlen(uri) == 8)
      sscanf(uri, "%x", &id);
   return id;
}


int readLn(char *buf, int max, int filedes)
{
   char c;
   int m, n;

   n = 0;
   do
   {
      m = recv(filedes, &c, 1, 0);
      if (m && (c != '\r'))
      {
         buf[n] = c;
         n++;
      }
   }
   while ((m) && (n < max) && (c != '\n') /* && (c != '\r') */ );
   buf[n] = '\0';
   return n + 1;
}

/*===========================================================================*/
/*==========                   Common Functions                    ==========*/
/*===========================================================================*/


struct_http_header *httpRecvHeader(int filedes)
{
   char *buf;
   struct_http_header *http_header;
   fd_set set;
   struct timeval timeout;
   int sel;
   char acceptGzip[256];
   char encodedGzip[256];

   FD_ZERO(&set);
   FD_SET(filedes, &set);
   timeout.tv_sec = 120;
   timeout.tv_usec = 0;
   sel = select(FD_SETSIZE, &set, NULL, NULL, &timeout);

   if (sel > 0)
   {
      buf = (char *) malloc(1024);
      http_header = calloc(sizeof(struct_http_header), 1);
      do
      {
         readLn(buf, 1024, filedes);
         while ((buf[strlen(buf) - 1] == '\n') || (buf[strlen(buf) - 1] == '\r'))
         {
            buf[strlen(buf) - 1] = '\0';
         }
         if (sscanf(buf, "GET %s HTTP/1.%*c", http_header->uri) > 0)
            http_header->header_type = HTTP_GET;
         else if (sscanf(buf, "POST %s HTTP/1.%*c", http_header->uri) > 0)
            http_header->header_type = HTTP_POST;
         else
            sscanf(buf, "HTTP/1.%*c %i", &http_header->response_code);
         sscanf(buf, "%*content-%*cength: %i", &http_header->data_length);
         sscanf(buf, "%*content-%*cocation: %s", http_header->uri);
         sscanf(buf, "%*content-%*cype: %s", http_header->content_type);
         sscanf(buf, "%*content-MD5: %s", http_header->md5);
         sscanf(buf, "TE: %s", acceptGzip);
         sscanf(buf, "%*cransfer-%*cncoding: %s", encodedGzip);
      }
      while (strlen(buf) > 0);
      free(buf);

      buf = strstr(http_header->uri, "http://");
      if (buf)
         buf += 7;
      else
         buf = http_header->uri;
      if (strchr(buf, '/'))
         strcpy(http_header->uri, strchr(buf, '/') + 1);
      //printf("HTTP Header:\n  URI: %s\n  MD5: %s\n  Len: %i\n", http_header->uri, http_header->md5, http_header->data_length);
      if (!strcmp(acceptGzip, "gzip") || !strcmp(encodedGzip, "gzip"))
         http_header->gzip = 1;
      return http_header;
   } else
   {
      if (sel == 0)
         fprintf(stderr, "recvHttpHeader timed out\n");
      else
         perror("recvHttpHeader");
      fflush(stderr);
      return NULL;
   }
}


int httpRecvData(char *buf, int length, int filedes)
{
   int read_bytes, total_bytes, nbytes;
   fd_set set;
   struct timeval timeout;
   int sel;

   FD_ZERO(&set);
   FD_SET(filedes, &set);
   timeout.tv_sec = 60;
   timeout.tv_usec = 0;
   sel = select(FD_SETSIZE, &set, NULL, NULL, &timeout);

   if (sel > 0)
   {
      total_bytes = 0;

      if (length - total_bytes >= READBUFFER)
         nbytes = READBUFFER;
      else
         nbytes = length - total_bytes;
      do
      {
         read_bytes = recv(filedes, buf + total_bytes, nbytes, 0);
         total_bytes += read_bytes;
         if (length - total_bytes < READBUFFER)
            nbytes = length - total_bytes;
      }
      while (nbytes && read_bytes);

      buf[total_bytes] = '\0';
      return total_bytes;
   } else
   {
      if (sel == 0)
         fprintf(stderr, "httpRecvData timed out\n");
      else
         perror("httpRecvData");
      fflush(stderr);
      buf[0] = 0;
      return 0;
   }
}




/*===========================================================================*/
/*==========                   Server Functions                    ==========*/
/*===========================================================================*/
void setupHttpServer(int port)
{
   int i = 1;

#ifdef _WIN32
   WSADATA wsaData;
   WSAStartup(MAKEWORD(1, 1), &wsaData);
#endif

   FD_ZERO(&fd_listen_set);

   listen_sock = socket(PF_INET, SOCK_STREAM, 0);
   if (listen_sock < 0)
   {
      perror("socket");
      exit(EXIT_FAILURE);
   }
#ifdef _WIN32
   setsockopt(listen_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &i, sizeof(int));
#else
   setsockopt(listen_sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(int));
#endif

   listen_sock_name.sin_family = AF_INET;
   listen_sock_name.sin_port = htons(port);
   listen_sock_name.sin_addr.s_addr = htonl(INADDR_ANY);
   if (bind(listen_sock, (struct sockaddr *) &listen_sock_name, sizeof(listen_sock_name)) < 0)
   {
      perror("bind");
      exit(EXIT_FAILURE);
   }

   if (listen(listen_sock, MAX_SOCKS) < 0)
   {
      perror("listen");
      exit(EXIT_FAILURE);
   }
   FD_SET(listen_sock, &fd_listen_set);
}


int waitForSocket()
{
   struct sockaddr_in client_name;
   int size = sizeof(client_name);
   int new_sock = -1;
   int waiting_conns;

   waiting_conns = select(FD_SETSIZE, &fd_listen_set, NULL, NULL, NULL);
   if (waiting_conns > 0)
   {
      new_sock = accept(listen_sock, (struct sockaddr *) &client_name, &size);
//    printf("Connection from %s:%i\n", inet_ntoa(client_name.sin_addr), ntohs(client_name.sin_port));
      return new_sock;
   }
   return 0;
}

int closeSock(int sock)
{
   struct linger lin;

   lin.l_onoff = 1;
   lin.l_linger = 300;
#ifdef _WIN32
   setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &lin, sizeof(lin));
   if (closesocket(sock))
   {
      perror("closeSock");
      return -1;
   }
#else
   setsockopt(sock, SOL_SOCKET, SO_LINGER, &lin, sizeof(lin));
   if (close(sock))
   {
      perror("closeSock");
      return -1;
   }
#endif

   return 0;
}


/*===========================================================================*/
/*==========                   Client Functions                    ==========*/
/*===========================================================================*/
int openConn(const char *host_name, int port)
{
   struct sockaddr_in dest;
   struct hostent *destInfo;
   int sock;

#ifdef _WIN32
   WSADATA wsaData;
   WSAStartup(MAKEWORD(1, 1), &wsaData);
#endif

   dest.sin_family = AF_INET;
   dest.sin_port = htons(port);
   destInfo = gethostbyname(host_name);

   if (destInfo == NULL)
   {
      fprintf(stderr, "Unknown host %s.\n", host_name);
      exit(EXIT_FAILURE);
   }
   dest.sin_addr = *(struct in_addr *) destInfo->h_addr;
   sock = socket(PF_INET, SOCK_STREAM, 0);
   if (sock < 0)
   {
      perror("openConn (socket)");
      return -1;
   }
   if (connect(sock, (struct sockaddr *) &dest, sizeof(dest)) < 0)
   {
      perror("openConn (connect)");
      close(sock);
      return -1;
   }
   return sock;
}




/*===========================================================================*/
/*==========                    Comm Interface                     ==========*/
/*===========================================================================*/

int sendHttpFile(struct_http_response * http_response)
{
   char *buf;
   char *md5;
   int read_bytes, written_bytes, total_bytes, nbytes;
   int file, sock;
   char *file_name;
#ifdef GZIP
   gzFile gzFd;
#endif

   nbytes = READBUFFER;
   buf = (char *) malloc(nbytes);
   sock = http_response->sock;

   file = open(http_response->file_name, O_RDONLY);
   if (file < 0)
   {
      sprintf(buf, "HTTP/1.1 404 %s\nConnection: close\nCache-Control: no-cache\nContent-Length: 0\n\n", "ERROR" /*strerror(errno) */ );
      send(sock, buf, strlen(buf), 0);
   } else
   {
      md5 = getMD5(file);
      file_name = strrchr(http_response->file_name, '/');
      if (file_name)
         file_name++;
      else
         file_name = http_response->file_name;

      if (http_response->gzip)
         sprintf(buf, "HTTP/1.1 %03i %s\nTransfer-Encoding: gzip\nConnection: close\nContent-Location: %s\nCache-Control: no-cache\nContent-MD5: %s\n\n", http_response->response_code, http_response->response_text, file_name, md5);
      else
      {
         http_response->length = (int) getFileSize(file);
         sprintf(buf, "HTTP/1.1 %03i %s\nConnection: close\nContent-Location: %s\nCache-Control: no-cache\nContent-Length: %i\nContent-MD5: %s\n\n", http_response->response_code, http_response->response_text, file_name, http_response->length, md5);
      }
      send(sock, buf, strlen(buf), 0);
      free(md5);

      lseek(file, 0, SEEK_SET);
      total_bytes = 0;
      if (http_response->gzip)
      {
#ifdef GZIP
         gzFd = gzdopen(dup(sock), "wb");
         do
         {
            read_bytes = read(file, buf, nbytes);
            if (read_bytes < 0)
               perror("sendHttpFile: read");
            written_bytes = gzwrite(gzFd, buf, read_bytes);
            if (written_bytes != read_bytes)
               perror("sendHttpFile: write");

            total_bytes += read_bytes;
         }
         while (read_bytes);
         gzclose(gzFd);
#endif
      } else
      {
         do
         {
            read_bytes = read(file, buf, nbytes);
            written_bytes = send(sock, buf, read_bytes, 0);
            if (written_bytes != read_bytes)
               perror("sendHttpFile: send");

            total_bytes += read_bytes;
            if (http_response->length - total_bytes < READBUFFER)
               nbytes = http_response->length - total_bytes;
         }
         while (read_bytes);
      }
      shutdown(sock, 1);
      closeSock(sock);
      close(file);
   }
   free(buf);
   free(http_response->buf);
   free(http_response);
   http_response = NULL;

   return 0;
}









/*===========================================================================*/
/*==========                   Server Functions                    ==========*/
/*===========================================================================*/

struct_http_response *createNewHttpResponse(int response_code, char *response_text, int gzip, char *file_name, int sock)
{
   struct_http_response *http_response;
   http_response = (struct_http_response *) malloc(sizeof(struct_http_response));

   http_response->response_code = response_code;
   strncpy(http_response->response_text, response_text, 255);
   http_response->gzip = gzip;
   strncpy(http_response->file_name, file_name, 255);
   http_response->sock = sock;
   http_response->length = 0;
   http_response->buf_size = 16384;
   http_response->buf = (char *) malloc(http_response->buf_size);
   http_response->buf[0] = '\0';
   return http_response;
}


int appendHttpResponse(struct_http_response * http_response, char *data, int len)
{
   char *new_buf;
   if (http_response->length + len > http_response->buf_size)
   {
      new_buf = (char *) realloc(http_response->buf, http_response->buf_size + len + 16384);
      if (new_buf)
      {
         http_response->buf = new_buf;
         http_response->buf_size += len + 16384;
      } else
      {
         fprintf(stderr, "appendHttpResponse: Unable to increase buffer, no memory available !\n");
         return -1;
      }
   }
   memcpy(http_response->buf + http_response->length, data, len);
   http_response->length += len;
   return 0;
}


int sendHttpResponse(struct_http_response * http_response)
{
   char *buf;
   char md5[33];
   struct MD5Context *context;
   unsigned char digest[16];
   int sock;
   int i;
   int total_bytes, written_bytes, nbytes;

   sock = http_response->sock;

   if (strlen(http_response->file_name))
   {
      sendHttpFile(http_response);
   } else
   {
      buf = (char *) malloc(2048);
      if (http_response->length)
      {
         context = malloc(sizeof(struct MD5Context));
         MD5Init(context);
         MD5Update(context, http_response->buf, http_response->length);
         MD5Final(digest, context);
         free(context);
         for (i = 0; i < 16; i++)
            sprintf(md5 + i * 2, "%02x", digest[i]);
         md5[32] = '\0';
         sprintf(buf, "HTTP/1.1 %03i %s\nConnection: close\nCache-Control: no-cache\nContent-Length: %i\nContent-MD5: %s\n\n", http_response->response_code, http_response->response_text, http_response->length, md5);
         send(sock, buf, strlen(buf), 0);
         total_bytes = 0;
         if (http_response->length - total_bytes >= READBUFFER)
            nbytes = READBUFFER;
         else
            nbytes = http_response->length - total_bytes;
         do
         {
            written_bytes = send(sock, http_response->buf + total_bytes, nbytes, 0);
            if (written_bytes != nbytes)
            {
               perror("sendHttpResponse (send)");
            }

            total_bytes += written_bytes;
            if (http_response->length - total_bytes < READBUFFER)
               nbytes = http_response->length - total_bytes;
         }
         while (total_bytes < http_response->length);
      } else
      {
         sprintf(buf, "HTTP/1.1 %03i %s\nConnection: close\nCache-Control: no-cache\nContent-Length: %i\n\n", http_response->response_code, http_response->response_text, http_response->length);
         send(sock, buf, strlen(buf), 0);
      }

      shutdown(sock, 1);
      closeSock(sock);
      free(buf);
      free(http_response->buf);
      free(http_response);
      http_response = NULL;
   }
   return 0;
}


int recvHttpReq(char **msg, struct_http_header ** httpHeader)
{
   int sock;
   struct_http_header *http_header;
   int i;
   int ret;

   http_header = NULL;
   sock = waitForSocket();
   ret = -1;
   i = 0;

   *msg = NULL;
   if (sock >= 0)
   {
      http_header = httpRecvHeader(sock);
      if (http_header)
      {
         ret = sock;
         *httpHeader = http_header;
         if (http_header->header_type == HTTP_POST)
         {
            if (http_header->data_length > 0)
            {
               *msg = (char *) malloc(http_header->data_length + 1);
               i = httpRecvData(*msg, http_header->data_length, sock);
               if (i != http_header->data_length)
               {
                  fprintf(stderr, "recvHttpReq: Only received %d of %d bytes\n", i, http_header->data_length);
                  free(*msg);
                  *msg = NULL;
                  ret = -1;
                  shutdown(sock, 1);
                  closeSock(sock);
               }
            }
         }
      }
   }
   return ret;
}






/*===========================================================================*/
/*==========                   Client Functions                    ==========*/
/*===========================================================================*/

void setupHttpProxy(char *host, int port)
{
   if (host && port)
   {
      proxy_host = strdup(host);
      proxy_port = port;
   }
}


struct_http_req *createNewHttpReq(int type, char *host, int port, char *uri, int gzip)
{
   struct_http_req *http_req;
   http_req = (struct_http_req *) malloc(sizeof(struct_http_req));
   http_req->type = type;
   strncpy(http_req->host, host, 255);
   http_req->port = port;
   if (uri[0] != '/')
      strcpy(http_req->uri, "/");
   else
      http_req->uri[0] = '\0';
   strncat(http_req->uri, uri, 255);
   http_req->length = 0;
   http_req->gzip = gzip;
   http_req->buf_size = 16384;
   http_req->buf = (char *) malloc(http_req->buf_size);
   http_req->buf[0] = '\0';
   return http_req;
}


int appendHttpReq(struct_http_req * http_req, char *data, int len)
{
   char *new_buf;
   if (http_req->length + len > http_req->buf_size)
   {
      new_buf = (char *) realloc(http_req->buf, http_req->buf_size + len + 16384);
      if (new_buf)
      {
         http_req->buf = new_buf;
         http_req->buf_size += 16384 + len;
      } else
      {
         fprintf(stderr, "appendHttpReq: Unable to increase buffer, no memory available !\n");
         return -1;
      }
   }
   memcpy(http_req->buf + http_req->length, data, len);
   http_req->length += len;
   return 0;
}


char *sendHttpReq(struct_http_req * http_req, int *status)
{
   char *buf;
   char md5[33];
   int sock, file;
   int n;
   struct_http_header *http_header;
   struct MD5Context *context;
   unsigned char digest[16];
   int total_bytes = 0;
   int read_bytes = 0;
   int written_bytes = 0;
   int nbytes = READBUFFER;
   char *file_name;
   int i;
   int eCode;
#ifdef GZIP
   char *acceptGzip = "TE: gzip\n";
   gzFile gzFd;
#else
   char *acceptGzip = "";
#endif

   if (proxy_host && proxy_port)
      sock = openConn(proxy_host, proxy_port);
   else
      sock = openConn(http_req->host, http_req->port);

   buf = NULL;
   eCode = 0;
   if (sock >= 0)
   {
      buf = (char *) malloc(2048);

      if (http_req->type == HTTP_GET)
      {
#ifdef GZIP
         printf("Request gzip-Encoding\n");
#endif
         sprintf(buf, "GET http://%s:%i%s HTTP/1.1\nHost: %s:%i\n%sContent-Length: %i\n\n", http_req->host, http_req->port, http_req->uri, http_req->host, http_req->port, acceptGzip, http_req->length);
      }
      if (http_req->type == HTTP_POST)
      {
         context = malloc(sizeof(struct MD5Context));
         MD5Init(context);
         MD5Update(context, http_req->buf, http_req->length);
         MD5Final(digest, context);
         free(context);
         for (i = 0; i < 16; i++)
            sprintf(md5 + i * 2, "%02x", digest[i]);
         md5[32] = '\0';
         sprintf(buf, "POST http://%s:%i%s HTTP/1.1\nHost: %s:%i\nContent-Length: %i\nContent-MD5: %s\n\n", http_req->host, http_req->port, http_req->uri, http_req->host, http_req->port, http_req->length, md5);
      }

      send(sock, buf, strlen(buf), 0);
      if (http_req->length)
         send(sock, http_req->buf, http_req->length, 0);
      shutdown(sock, 1);
      free(buf);
      free(http_req->buf);
      free(http_req);
      http_req = NULL;

      buf = NULL;
      http_header = httpRecvHeader(sock);
      if ((http_header) && (http_header->response_code >= 200) && (http_header->response_code < 300))
      {
         if (http_header->response_code == 200)
         {
            buf = (char *) malloc(READBUFFER);
            context = malloc(sizeof(struct MD5Context));
            file_name = strrchr(http_header->uri, '/');
            if (file_name != NULL)
               file_name += 1;
            else
               file_name = http_header->uri;

            file = open(file_name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
            MD5Init(context);
            nbytes = READBUFFER;

            if (http_header->gzip)
            {
               printf("Data is obviously gzip-encoded\n");
#ifdef GZIP
               gzFd = gzdopen(dup(sock), "rb");
               do
               {
                  read_bytes = gzread(gzFd, buf, nbytes);
                  MD5Update(context, buf, read_bytes);
                  written_bytes = write(file, buf, read_bytes);
                  if (written_bytes != read_bytes)
                  {
                     perror("write");
                     *status = -255;
                     free(buf);
                     return NULL;
                  }
                  total_bytes += read_bytes;
               }
               while (read_bytes);
               gzclose(gzFd);
#endif
            } else
            {
               do
               {
                  read_bytes = recv(sock, buf, nbytes, 0);
                  MD5Update(context, buf, read_bytes);
                  written_bytes = write(file, buf, read_bytes);
                  if (written_bytes != read_bytes)
                  {
                     perror("write");
                     *status = -255;
                     free(buf);
                     return NULL;
                  }
                  total_bytes += read_bytes;
               }
               while (read_bytes);
            }

            free(buf);
            close(file);
            MD5Final(digest, context);
            free(context);
            for (i = 0; i < 16; i++)
               sprintf(md5 + i * 2, "%02x", digest[i]);
            md5[32] = '\0';
            if (strcmp(md5, http_header->md5))
            {
               fprintf(stderr, "Error while receiving file: MD5-Error\n");
               buf = NULL;
               eCode = -1;
            } else
               buf = strdup(file_name);
         } else
         {
            if (http_header->data_length > 0)
            {
               buf = malloc(http_header->data_length + 1);
               n = httpRecvData(buf, http_header->data_length, sock);
               if (n != http_header->data_length)
               {
                  fprintf(stderr, "sendHttpReq: Only received %d of %d bytes\n", n, http_header->data_length);
                  free(buf);
                  buf = NULL;
                  eCode = -1;
               }
            }
         }
         free(http_header);
      } else
      {
         if (http_header)
         {
            fprintf(stderr, "sendHttpReq: Master reported error: %d\n", http_header->response_code);
            free(http_header);
         } else
            fprintf(stderr, "We did not receive a answer");
         fflush(stderr);
         eCode = -2;
      }
      closeSock(sock);
   } else
   {
      perror("sendHttpReq: Unable to open socket");
      eCode = -3;
   }
   *status = eCode;
   return buf;
}
