#include "axml.h"
#include "common.h"
#include "http_comm.h"


int myNum, workers;
double globTime = 0.0;
int permutationSendCount = 0;
int permutationRecvCount = 0;
int numOfPerm = 20;
int masterPort = 5555;
int numOfWorkers = 0;
int worker_count = 0;


int getTreeIds(int **treeIds, char **msg)
{
   char *s1, *s2, *s3;
   char *s;
   char *temp;
   int id_ptr, id, i;
   s = (char *) malloc(20);

   id_ptr = -1;
   s3 = NULL;

   if (s1 = strstr(*msg, "}#{"))
      if (s2 = strstr(s1 + 2, "}#{"))
         s3 = strstr(s2 + 2, "}#{");

   if (s3)
   {
      s1 = strchr(*msg, '{');
      if (s1)
      {
         s2 = strstr(s1, "}#{");
         s3 = s2;
         id_ptr = 1;
         *treeIds = (int *) malloc((((s2 - s1) / 9) + 5) * sizeof(int));
         while ((s2 = strchr(++s1, ' ')) && (s2 < s3))
         {
            strncpy(s, s1, s2 - s1);
            s[s2 - s1] = 0;
            sscanf(s, "%08x", &id);
            (*treeIds)[id_ptr++] = id;
            s1 = s2;
         }
         strncpy(s, s1, s3 - s1);
         sscanf(s, "%08x", &id);
         (*treeIds)[id_ptr++] = id;
         s1 = strdup(s3 + 2);
         free(*msg);
         *msg = s1;
         (*treeIds)[0] = id_ptr - 1;
      }
   } else
      *treeIds = NULL;
   free(s);
   return id_ptr - 1;
}


int handleHttpReq(int *work, int *jobId, char *tempStr, int **treeIds)
{
   int sock;
   int worker_id;
   int id_count;
   struct_http_header *http_header;
   struct_http_response *http_response;
   char buf[32];
   char *msg;

   sock = recvHttpReq(&msg, &http_header);
   if (http_header->header_type == HTTP_GET)
   {
      if (!strcmp(http_header->uri, "REGISTER"))
      {
         worker_id = time(NULL);
         sprintf(buf, "%08x", worker_id);
         http_response = createNewHttpResponse(202, "Accepted", 0, "", sock);
         appendHttpResponse(http_response, buf, strlen(buf));
         worker_count++;
      } else
         http_response = createNewHttpResponse(200, "OK", http_header->gzip, seq_file, sock);
      sendHttpResponse(http_response);
      free(msg);
      return -1;
   } else
   {
      id_count = getTreeIds(treeIds, &msg);
      getTreeStr(msg, work, jobId, tempStr);
      return sock;
   }
}

int sendHttpAnswer(char *buf, int sock)
{
   pthread_t thread_id;
   struct_http_response *http_response;

   http_response = createNewHttpResponse(202, "Accepted", 0, "", sock);
   if (buf)
      appendHttpResponse(http_response, buf, strlen(buf));

   pthread_create(&thread_id, NULL, sendHttpResponse, (void *) http_response);
}


int get_args(int argc, char *argv[], boolean print_usage, analdef * adef)
{                               /* get_args */
   int c;
   char *s;
   boolean bad_opt = FALSE;
   int i, time;
   char buf[2048];


   run_id[0] = 0;
   workdir[0] = 0;
   seq_file[0] = 0;
   tree_file[0] = 0;
   adef->restart = FALSE;


   while (!bad_opt && ((c = getopt(argc, argv, "n:w:o:t:p:x:r:")) != -1))
   {
      /*fprintf(stderr, "%s\n", optarg); */
      switch (c)
      {
      case 'n':                /* Name, or ID of run */
         strcpy(run_id, optarg);
         break;
      case 'w':                /* Working directory */
         strcpy(workdir, optarg);
         break;
      case 'o':
         sscanf(optarg, "%d", &masterPort);
         printf("  Port: %d\n", masterPort);
         break;
      case 't':
         strcpy(tree_file, optarg);
         adef->restart = TRUE;
         break;
      case 'p':
         sscanf(optarg, "%d", &numOfPerm);
         printf("  numOfPerm: %d\n", numOfPerm);
         break;
      case 'x':
         sscanf(optarg, "%d", &numOfWorkers);
         printf("  numOfWorkers: %d\n", numOfWorkers);
         break;
      case 'r':
         sscanf(optarg, "%d", &rearrangementStep);
         if (rearrangementStep < 2)
            rearrangementStep = 2;
         break;
      case '?':
      default:
         bad_opt = TRUE;
         break;
      }
   }


/* Set the full path to the working directory. It's up to using functions
 * of workdir to resolve this path, which may contain instances of ./ ../
 * and //, and handle resulting errors. (N.B.: We take care not to introduce
 * double slashes because some implementations of realpath() (notably in AIX)
 * choke on them). */
   if (workdir[0] == 0 || workdir[0] != '/')
   {
      getcwd(buf, sizeof(buf));
      if (buf[strlen(buf) - 1] != '/')
         strcat(buf, "/");
      strcat(buf, workdir);
      if (buf[strlen(buf) - 1] != '/')
         strcat(buf, "/");
      strcpy(workdir, buf);
   }

/* There should be one argument left, the sequence data file name. */
   if (optind == argc - 1)
   {
      strcpy(seq_file, argv[argc - 1]);
   }

/* Check for errors */
   if (bad_opt)
   {
      if (print_usage)
      {
         s = strrchr(argv[0], '/');
         s = (s == NULL) ? argv[0] : s + 1;
         fprintf(stderr, "Usage: %s [-n <runName>] [-w <workDir>] [-o <port>] [-t <treeFile>] [-p <numOfPerm>] [-x <numOfWorkers>] [-r <rearrange>] <sequenceFile>\n", s);
         fflush(stderr);
      }
      return 0;
   } else
   {
      if (run_id[0] == 0)
      {
         s = strrchr(seq_file, '/');
         s = (s == NULL) ? seq_file : s + 1;
         strcpy(run_id, s);
      }
      return 1;
   }
}                               /* get_args */




/*==========================================================================*/
/*                             "main" routine                               */
/*==========================================================================*/


int main(int argc, char *argv[])
{                               /* DNA Maximum Likelihood */
   analdef *adef;
   rawdata *rdta;
   cruncheddata *cdta;
   tree *tr;                    /*  current tree */

   int j, *perm, l;
   int myprogtype;
   int *prog;
   double workerRLH;
   FILE *permutationFile, *resultFile;
   double masterTime;
   char permFileName[1024], resultFileName[1024];
   int parsimonyTrees = 0;
   int ptreesCount = 0;
   int endIt = 0;
   int work[10];
   int rearrangementsMax, rearrangementsMin;
   char *treeVector;
   int numOfOptimizeJobs, optimizeJobsSent, optimizeJobsReceived, numOfEvaluateJobs, evaluateJobsSent, evaluateJobsReceived;
   int from;
   double lh, recvdlh, optlh, t;
   int i, k, impr, stepwidth;
   jobQueue *jobList;
   int betterFound;
   char *parsimonyTree;

   int sock;
   int jobId;
   char *msg;
   char *treeBuf;
   char *tempStr;
   char *tmp;
   int curJobId;
   int curJobIdBase;
   int resendCount;
   double testLH;
   double testTime;
   int *treeIds;
   int allReturned;
   char textOutFileName[1024];
   FILE *treeOutFile;
   char treeOutFileName[1024];

   printf("%s %s with %s %s (%s)\n\n", programName, programVersion, http_comm_name, http_comm_version, http_comm_date);

   CATEGORIES = 0;
   Multiple = 1;
   plist = NULL;
   rearrangementStep = 5;

   curJobId = 0;
   workers = 1;

   if (!(adef = (analdef *) malloc(sizeof(analdef))))
   {
      printf("ERROR: Unable to get memory for analysis definition\n\n");
      return 1;
   }

   if (!get_args(argc, argv, (boolean) 1, adef));

   if (!(rdta = (rawdata *) malloc(sizeof(rawdata))))
   {
      printf("ERROR: Unable to get memory for raw DNA\n\n");
      return 1;
   }

   if (!(cdta = (cruncheddata *) malloc(sizeof(cruncheddata))))
   {
      printf("ERROR: Unable to get memory for crunched DNA\n\n");
      return 1;
   }


   tr = (tree *) malloc(sizeof(tree));


   while (1)
   {
      printf("\nEntering main loop\n");
      masterTime = gettime();

      INFILE = fopen(seq_file, "r");
      printf("Sequence-File: %s\n", seq_file);
      if (!getinput(adef, rdta, cdta, tr))
         return 1;
      fclose(INFILE);

      if (!makeweights(adef, rdta, cdta))
         return 1;
      if (!makevalues(rdta, cdta))
         return 1;
      if (adef->empf && !empiricalfreqs(rdta, cdta))
         return 1;
      reportfreqs(adef, rdta);
      if (!linkdata2tree(rdta, cdta, tr))
         return 1;
      if (!linkxarray(3, 3, cdta->endsite, &freextip, &usedxtip))
         return 1;
      if (!setupnodex(tr))
         return 1;
      if (!setupEqualityVector(tr))
         return 1;
      treeStringLength = (tr->mxtips * (nmlngth + 32)) + 256 + (tr->mxtips * 2);
      myTreestr = (char *) malloc(treeStringLength);
      bestTreestr = (char *) malloc(treeStringLength);
      perm = (int *) malloc((tr->mxtips + 1) * sizeof(int));
      tl = (treelist *) malloc(sizeof(treelist));
      tempStr = (char *) malloc(treeStringLength);
      jobList = (jobQueue *) malloc(sizeof(jobQueue) * 2 * tr->mxtips);

      initList(20);

      strcpy(permFileName, workdir);
      strcpy(resultFileName, workdir);
      strcpy(textOutFileName, workdir);
      strcpy(treeOutFileName, workdir);

      strcat(permFileName, "RAxML_perm.");
      strcat(resultFileName, "RAxML_result.");
      strcat(textOutFileName, "RAxML_stderr.");
      strcat(treeOutFileName, "RAxML_tree.");

      strcat(permFileName, run_id);
      strcat(resultFileName, run_id);
      strcat(textOutFileName, run_id);
      strcat(treeOutFileName, run_id);

      setupHttpServer(masterPort);

      while (1)
      {

         if (adef->restart)
         {
            INFILE = fopen(tree_file, "r");
            if (!treeReadLen(INFILE, tr))
               return FALSE;
            if (!smoothTree(tr, smoothings))
               return FALSE;
            if (CATEGORIES)
            {
               if (evaluateCAT(tr, tr->start) == badEval)
                  return FALSE;
            } else
            {
               if (evaluate(tr, tr->start) == badEval)
                  return FALSE;
            }
            lh = tr->likelihood;
            treeString(myTreestr, tr, tr->start->back, 1);
            strcpy(bestTreestr, myTreestr);
            plist = permutationListInsert(lh, myTreestr, plist, tr);
            fclose(INFILE);
         } else
         {
            printf("Entering phase #1\n");
            while (permutationRecvCount < numOfPerm)
            {
               sock = handleHttpReq(work, &jobId, myTreestr, &treeIds);
               if (sock >= 0)
               {
                  switch (work[8])
                  {
                  case SIMPLE_REQUEST:
                     if (permutationRecvCount < numOfPerm)
                     {
                        work[8] = RANDOM_TREE;
                        treeBuf = setTreeStr(work, curJobId++, NULL);
                        sendHttpAnswer(treeBuf, sock);
                        free(treeBuf);
                        permutationSendCount++;
                     }
                     break;
                  case PARSIMONY_TREE:
                     workerRLH = str_readTreeLikelihood(myTreestr);
                     testTime = gettime();
                     str_treeReadLen(myTreestr, tr);
                     if (CATEGORIES)
                        testLH = evaluateCAT(tr, tr->start);
                     else
                        testLH = evaluate(tr, tr->start);
                     if (workerRLH > testLH * 1.01 && workerRLH < testLH * 0.99)
                     {
                        permutationRecvCount++;
                        plist = permutationListInsert(workerRLH, myTreestr, plist, tr);
                        permutationFile = fopen(permFileName, "w");
                        printList(plist, permutationFile);
                        fclose(permutationFile);
                     } else
                        fprintf(stderr, "Likelihood could not be verified (reported/verified): %f/%f\n", workerRLH, testLH);
                     sendHttpAnswer(NULL, sock);
                     break;
                  }
               }
            }
         }

         printf("Entering phase #2\n");
         fprintf(stderr, "%d/%d Permutationjobs sent/received\n", permutationSendCount, permutationRecvCount);

         str_treeReadLen(getList(plist, 0), tr);
         treeOutFile = fopen(treeOutFileName, "w");
         treeOut(treeOutFile, tr, 1);
         fclose(treeOutFile);

         impr = 1;

         while (1)
         {
            fflush(stderr);
            t = gettime();
            curJobId = (curJobId & 0xF0000000) + 0x10000000;
            curJobIdBase = curJobId;
            if (numOfWorkers)
               workers = numOfWorkers;
            else
               workers = worker_count;

            if (workers > 3)
               stepwidth = (2 * tr->mxtips - 1) / (workers * 2);
            else
               stepwidth = 10;
            numOfOptimizeJobs = (2 * tr->mxtips - 1) / stepwidth;

            optimizeJobsSent = 0;
            optimizeJobsReceived = 0;

            numOfEvaluateJobs = workers;
            evaluateJobsSent = 0;
            evaluateJobsReceived = 0;

            jobList[0].lower = 2;
            jobList[0].upper = 2 + stepwidth;

            for (i = 1; i < numOfOptimizeJobs; i++)
            {
               jobList[i].lower = jobList[i - 1].upper;
               jobList[i].upper = jobList[i].lower + stepwidth;
            }

            jobList[numOfOptimizeJobs - 1].upper = 2 * tr->mxtips - 1;

            for (i = 0; i < numOfOptimizeJobs; i++)
            {
               jobList[i].jobId = curJobId++;
               jobList[i].gotOptResult = 0;
               jobList[i].gotTreeResult = 0;
            }

            str_treeReadLen(getList(plist, 0), tr);
            lh = tr->likelihood;
            treeString(myTreestr, tr, tr->start->back, 1);
            fprintf(stderr, "%f %f\n", gettime() - masterTime, lh);

            if (impr)
            {
               rearrangementsMax = rearrangementStep;
               rearrangementsMin = 1;
               resultFile = fopen(resultFileName, "w");
               fprintf(resultFile, "Time:%f, %s\n", gettime() - masterTime, myTreestr);
               fclose(resultFile);
            } else
            {
               if (Multiple == 1)
               {
                  Multiple = 0;
                  rearrangementsMax = rearrangementStep;
                  rearrangementsMin = 1;
               } else
               {
                  rearrangementsMax += rearrangementStep;
                  rearrangementsMin += rearrangementStep;
               }
               if (rearrangementsMax >= MAX_REARRANGE)
                  goto cleanup;
            }

            betterFound = 0;
            resendCount = 0;

            while (optimizeJobsReceived < numOfOptimizeJobs)
            {
               sock = handleHttpReq(work, &jobId, tempStr, &treeIds);
               if (sock >= 0)
               {
                  switch (work[8])
                  {
                  case SIMPLE_REQUEST:
                     work[0] = rearrangementsMin;
                     work[1] = rearrangementsMax;
                     work[2] = Multiple;
                     work[8] = TREE_OPTIMIZE;
                     if (optimizeJobsSent < numOfOptimizeJobs)
                     {
                        work[3] = jobList[optimizeJobsSent].lower;
                        work[4] = jobList[optimizeJobsSent].upper;
                        if (betterFound == 1 && Multiple == 1)
                           treeBuf = setTreeStr(work, jobList[optimizeJobsSent].jobId, bestTreestr);
                        else
                           treeBuf = setTreeStr(work, jobList[optimizeJobsSent].jobId, myTreestr);
                     } else
                     {
                        while (jobList[resendCount].gotOptResult)
                        {
                           resendCount++;
                           if (resendCount >= numOfOptimizeJobs)
                              resendCount = 0;
                        }
                        if (!jobList[resendCount].gotOptResult)
                        {
                           work[3] = jobList[resendCount].lower;
                           work[4] = jobList[resendCount].upper;
                           if (betterFound == 1 && Multiple == 1)
                              treeBuf = setTreeStr(work, jobList[resendCount].jobId, bestTreestr);
                           else
                              treeBuf = setTreeStr(work, jobList[resendCount].jobId, myTreestr);
                        }
                     }
                     sendHttpAnswer(treeBuf, sock);
                     free(treeBuf);
                     optimizeJobsSent++;
                     break;
                  case OPTIMIZE_REQUEST:
                     if (Multiple == 1)
                     {
                        if (str_readTreeLikelihood(tempStr) > str_readTreeLikelihood(myTreestr))
                        {
                           testTime = gettime();
                           str_treeReadLen(tempStr, tr);
                           if (CATEGORIES)
                              testLH = evaluateCAT(tr, tr->start);
                           else
                              testLH = evaluate(tr, tr->start);
                           workerRLH = str_readTreeLikelihood(tempStr);
                           if (workerRLH > testLH * 1.01 && workerRLH < testLH * 0.99)
                           {
                              betterFound = 1;
                              strcpy(bestTreestr, tempStr);
                           } else
                              fprintf(stderr, "Likelihood could not be verified (reported/tested): %f/%f\n", workerRLH, testLH);
                        }
                     }
                     sendHttpAnswer(NULL, sock);
                     if ((jobId >= curJobIdBase) && (jobId < curJobIdBase + numOfOptimizeJobs) && (!jobList[jobId - curJobIdBase].gotOptResult))
                     {
                        jobList[jobId - curJobIdBase].gotOptResult = 1;
                        optimizeJobsReceived++;
                     }
                     break;
                  default:
                     sendHttpAnswer(NULL, sock);
                     break;
                  }
               }
            }

            impr = 0;
            optimizeJobsReceived = 0;
            fflush(stderr);
            while (optimizeJobsReceived < numOfOptimizeJobs)
            {
               sock = handleHttpReq(work, &jobId, tempStr, &treeIds);
               if (sock >= 0)
               {
                  switch (work[8])
                  {
                  case SIMPLE_REQUEST:
                     if (evaluateJobsSent < numOfEvaluateJobs)
                     {
                        work[8] = TREE_EVALUATE;
                        treeBuf = setTreeStr(work, curJobId++, NULL);
                        evaluateJobsSent++;
                        sendHttpAnswer(treeBuf, sock);
                        free(treeBuf);
                     } else
                     {
                        work[0] = rearrangementsMin;
                        work[1] = rearrangementsMax;
                        work[2] = Multiple;
                        work[8] = COMBINED_OPTIMIZE_EVALUATE;
                        treeBuf = setTreeStr(work, curJobId++, bestTreestr);
                        resendCount = 0;
                        for (i = 0; i < numOfOptimizeJobs; i++)
                           if (!jobList[i].gotTreeResult)
                              resendCount++;
                        tmp = (char *) malloc(resendCount * 19 + strlen(treeBuf) + 128);
                        sprintf(tmp, "{%d ", resendCount);
                        for (i = 0; i < numOfOptimizeJobs; i++)
                           if (!jobList[i].gotTreeResult)
                           {
                              sprintf(tempStr, "(%x %x)", jobList[i].lower, jobList[i].upper);
                              strcat(tmp, tempStr);
                           }
                        strcat(tmp, "}#");
                        strcat(tmp, treeBuf);
                        free(treeBuf);
                        sendHttpAnswer(tmp, sock);
                        free(tmp);
                        evaluateJobsSent++;
                     }
                     break;
                  case TREE_RESULT:
                     if (treeIds)
                     {
                        if (str_readTreeLikelihood(tempStr) > str_readTreeLikelihood(myTreestr))
                        {
                           testTime = gettime();
                           str_treeReadLen(tempStr, tr);
                           if (CATEGORIES)
                              testLH = evaluateCAT(tr, tr->start);
                           else
                              testLH = evaluate(tr, tr->start);
                           workerRLH = str_readTreeLikelihood(tempStr);
                           if (workerRLH > testLH * 1.01 && workerRLH < testLH * 0.99)
                           {
                              strcpy(myTreestr, tempStr);
                           } else
                              fprintf(stderr, "Likelihood could not be verified (reported/tested): %f/%f\n", workerRLH, testLH);
                           if ((recvdlh = str_readTreeLikelihood(myTreestr)) > lh)
                           {
                              impr = 1;
                              lh = tr->likelihood;
                              plist = permutationListInsert(recvdlh, myTreestr, plist, tr);
                           }
                        }
                        for (i = 1; i <= treeIds[0]; i++)
                           if ((treeIds[i] >= curJobIdBase) && (treeIds[i] < curJobIdBase + numOfOptimizeJobs))
                           {
                              if (!jobList[treeIds[i] - curJobIdBase].gotTreeResult)
                              {
                                 optimizeJobsReceived++;
                                 jobList[treeIds[i] - curJobIdBase].gotTreeResult = 1;
                              }
                              evaluateJobsReceived++;
                           }
                     }
                     sendHttpAnswer(NULL, sock);
                     break;
                  case COMBINED_RESULT:
                     if ((jobId & 0xF0000000) == (curJobIdBase & 0xF0000000))
                     {
                        if (str_readTreeLikelihood(tempStr) > str_readTreeLikelihood(myTreestr))
                        {
                           testTime = gettime();
                           str_treeReadLen(tempStr, tr);
                           if (CATEGORIES)
                              testLH = evaluateCAT(tr, tr->start);
                           else
                              testLH = evaluate(tr, tr->start);
                           workerRLH = str_readTreeLikelihood(tempStr);
                           if (workerRLH > testLH * 1.01 && workerRLH < testLH * 0.99)
                              strcpy(myTreestr, tempStr);
                           else
                              fprintf(stderr, "Likelihood could not be verified (reported/tested): %f/%f\n", workerRLH, testLH);
                           if ((recvdlh = str_readTreeLikelihood(myTreestr)) > lh)
                           {
                              impr = 1;
                              lh = tr->likelihood;
                              plist = permutationListInsert(recvdlh, myTreestr, plist, tr);
                           }
                        }
                        optimizeJobsReceived = numOfOptimizeJobs;
                     }
                     sendHttpAnswer(NULL, sock);
                     break;
                  default:
                     sendHttpAnswer(NULL, sock);
                     break;
                  }
               }
            }
         }
      }

    cleanup:
      resultFile = fopen(resultFileName, "w");
      fprintf(resultFile, "Time:%f, %s\n", gettime() - masterTime, myTreestr);
      fclose(resultFile);

      RAxML_parsimony_cleanup();

      free(perm);
      free(myTreestr);
      free(rdta->y0);
      if (rdta->wgt != NULL)
         free(rdta->wgt);
      if (rdta->wgt2 != NULL)
         free(rdta->wgt2);
      if (rdta->sitecat != NULL)
         free(rdta->sitecat);

      if (cdta->alias != NULL)
         free(cdta->alias);
      if (cdta->aliaswgt != NULL)
         free(cdta->aliaswgt);
      if (cdta->patcat != NULL)
         free(cdta->patcat);
      if (cdta->patrat != NULL)
         free(cdta->patrat);
      if (cdta->wr != NULL)
         free(cdta->wr);
      if (cdta->wr2 != NULL)
         free(cdta->wr2);
      freextip = (xarray *) NULL;
      usedxtip = (xarray *) NULL;
      freeTree(tr);


      plist = permutationListReset(plist);

      fprintf(stderr, "execution time: %f\n\n", gettime() - masterTime);
      fprintf(stderr, "Permutation Results written to: %s\n", permFileName);
      fprintf(stderr, "Rearrangement Results written to: %s\n", resultFileName);
      fflush(stderr);
      printf("Shutdown workers...\n");
      while (1)
      {
         sock = handleHttpReq(work, &jobId, tempStr, &treeIds);
         if (sock >= 0)
         {
            work[8] = FINISH_AXML;
            treeBuf = setTreeStr(work, curJobId++, NULL);
            sendHttpAnswer(treeBuf, sock);
            free(treeBuf);
         }
      }
      goto finish;
   }

 finish:
   return 0;
}
